#     Copyright (C) 2006
#
#     This code is distributed under the terms and conditions of the
#     CCP4 licence agreement as `Part 1' (Annex 2) software.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
# CCP4i database visualiser - viewer.tcl
#
# Usage: wish viewer.tcl

# Wanjuan Yang and Peter Briggs Oct 06
#
#====================================================================

#CCP4i_cvs_Id $Id: viewer.tcl,v 1.56 2007/02/09 16:08:39 wy45 Exp $

######################################################################
#d_index_title CCP4i database visualiser 
#d_intro_title Client application for database handler
#d_intro This is an application which displays ccp4i job database(i.e. database.def file). It requires db handler running in order to access job database.

# source API
source [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbClientAPI.tcl]

# Hide the Tk window
if { [lsearch -exact [package names] tk] >= -1 } {
    wm withdraw .
}
#
#----------------------------------------------------------------
proc GetEnvPath { var } {
#----------------------------------------------------------------
  global system
  global env

  # Get environment variable - cope with whether input var has
  # leading '$' or not

  regsub {^\$} $var  {} var1

  # On NT change the separators to unix style /

  if { [regexp WINDOWS $system(OPSYS)] } {
    set status [catch { global env;
      regsub -all {\\} $env($var1) \/ path } nc ]
  } else {
    set status [catch {global env; set p1 $env($var1)} path ]
  }
  if { $status == 1 } {
    set path ""
    WarningMessage "Can not get environment variable for $var"
  }
  return $path
}

# Source startup.tcl appropriate for operating system

set system(OPSYS) [string toupper $tcl_platform(platform)]
source [file join $env(CCP4I_TOP) bin $system(OPSYS) startup.tcl]

# Edit the RUN_MODE to be the name of the current application
set system(RUN_MODE)            mtzviewer
set system(TK_VERSION)          $tk_version

source [file join $env(CCP4I_TOP) src system.tcl]
source [file join $env(CCP4I_TOP) src utils.tcl]
source [file join $env(CCP4I_TOP) src window.tcl]
source [SearchPath TOP etc types.def]
source [SearchPath TOP src exframe.tcl]
source [SearchPath TOP src varmenu.tcl]
source [SearchPath TOP src fileselect.tcl]
source [SearchPath TOP src fileviewer.tcl]
source [SearchPath TOP src util_windows.tcl]
source [SearchPath TOP src local.tcl]

# If there is an autoconfigure procedure for the system parameters
# and this opsys then run it
if { [llength [info proc autoconf_[subst $system(OPSYS)]_system]] > 0 } {
      autoconf_[subst $system(OPSYS)]_system system }

InitialiseDotCCP4
global configure
InitialisePreferences configure configure
InitialisePreferences preferences preferences
InitialisePreferences directories directories

# Show the Tk window
wm deiconify .


######################################################################
# Global parameters
######################################################################
#
# These are stored as elements in an array
set ccp4i_db_viz(CANVAS) {}
set ccp4i_db_viz(GRAPH) {}
set ccp4i_db_viz(JOB_LIST) {}
set ccp4i_db_viz(SELECTION) {}
set ccp4i_db_viz(LAST_SELECTION_LIST) {}
set ccp4i_db_viz(FORWARD_SELECTION_LIST) {}
set ccp4i_db_viz(MAXWIDTH) 600
set ccp4i_db_viz(MAXHEIGHT) 600
set ccp4i_db_viz(MINWIDTH) 300
set ccp4i_db_viz(MINHEIGHT) 300
set ccp4i_db_viz(BORDER) 10
set ccp4i_db_viz(PROJECT) {}
set ccp4i_db_viz(DIRECTION) TB
set ccp4i_db_viz(SHAPE) ellipse
set ccp4i_db_viz(COLOR) black
set ccp4i_db_viz(BACKGROUND) {}
set ccp4i_db_viz(FILLCOLOR) {}
set ccp4i_db_viz(FINISHED_FILTER) 0
set ccp4i_db_viz(LINKED_FILTER) 0
############################################################
# General display procedures
############################################################

#---------------------------------------------------------------
proc create_main_window { } {
#----------------------------------------------------------------
    # This sets up the widgets for the main display
    global ccp4i_db_viz
    # "configure is CCP4i's global configuration
    # array which contains font and colour information
    global configure

    # Set up the canvas, buttons etc
    set f  [frame .f]
    set f1 [frame $f.f1]
    set f2 [frame $f.f2]
    set finfo [frame .finfo]


    # set the window title
    wm title . "CCP4Idbviewer"
    ########################
    # Canvas and scrollbars
    ########################
    set c [canvas $f1.c]
    set scrollx [scrollbar $f1.scrollx -orient horizontal -command "$c xview"]
    $c configure -xscrollcommand "$scrollx set"
    set scrolly [scrollbar $f2.scrolly -orient vertical -command "$c yview"]
    $c configure -yscrollcommand "$scrolly set"
    # Store the name of the canvas for later
    set ccp4i_db_viz(CANVAS) $c
    
    ###########
    # Menu bar
    ###########
    set fmbar [frame .fmbar -background $configure(COLOUR_CONTRAST)]

    #"File" menu
    set bfile [ menubutton $fmbar.file -text "File" \
		    -background $configure(COLOUR_CONTRAST) \
		    -menu $fmbar.file.menu]
    set filemenu [ menu $bfile.menu -tearoff 0 ]

    # add menu
    $filemenu add command -label "Output graph" -command "output_graph"
    $filemenu add command -label "Output xml" -command "output_xml"
    $filemenu add command -label "Exit" -command  { clear } 
    
    pack $bfile -side left

    # "Project" menu
    set bproj [ menubutton $fmbar.project -text "Project" \
		    -background $configure(COLOUR_CONTRAST) \
		    -menu $fmbar.project.menu]
    set projectmenu [ menu $bproj.menu -tearoff 0 ]
   
    #$projectmenu add cascade -label "Load project" -menu $projectmenu.project
    #menu $projectmenu.project

    # get list of projects & populate the menu
    set projects [ ListProjects ]
    if { $projects != 0  } {
	foreach project $projects {
	    $projectmenu add radiobutton -variable currentproject -label $project -command "select_project $project"
	}
    } else { 
	WarningMessage "no projects loaded"
	exit
    }
    
 
    #$projectmenu add command -label "Exit" -command  { clear } 
    pack $bproj -side left

 # "View" menu
    #set bedit [ menubutton $fmbar.edit -text "View" \
		    -background $configure(COLOUR_CONTRAST) \
		    -menu $fmbar.edit.menu]
    #set editmenu [ menu  $bedit.menu -tearoff 0 ]
 
    #$editmenu add command -label "Show only selected jobs" \
	-command "render_selected_jobs"
    #$editmenu add command -label "Show all jobs" \
	-command "render_all_jobs"
    #$editmenu add separator
    #$editmenu add command -label "Clear selected" -command "clear_selection $c"
    #$editmenu add separator
    #$editmenu add cascade -label "Filter" -menu $editmenu.menu
    #menu $editmenu.menu 
    #$editmenu.menu add command -label "Show Finished jobs" -command "render_finished_jobs"
    #$editmenu.menu add command -label "Show Linked jobs" -command "render_linked_jobs"
    #pack $bedit -side left

    # "Filter" menu
    set bfilter [menubutton $fmbar.filter -text "Filter" \
		   -background $configure(COLOUR_CONTRAST) \
		   -menu $fmbar.filter.menu]
    set filtermenu [ menu $bfilter.menu -tearoff 0 ]
    $filtermenu add checkbutton -variable ccp4i_db_viz(FINISHED_FILTER) -label "Show Finished jobs" 
#-command "render_finished_jobs" 
    $filtermenu add checkbutton -variable ccp4i_db_viz(LINKED_FILTER) -label "Show Linked jobs"
# -command "render_linked_jobs"
    pack $bfilter -side left

  # "Appearance" menu
    set bview [menubutton $fmbar.view -text "Appearance" \
		   -background $configure(COLOUR_CONTRAST) \
		   -menu $fmbar.view.menu]
    set viewmenu [ menu $bview.menu -tearoff 0 ]
    
    # background 
    $viewmenu add cascade -label "Background" \
	-menu $viewmenu.bg
    menu $viewmenu.bg
    $viewmenu.bg add radiobutton -variable background -label "white" -command {set ccp4i_db_viz(BACKGROUND) white ; re_display}
    $viewmenu.bg add radiobutton -label "gray" -variable background -command {set ccp4i_db_viz(BACKGROUND) grey ; re_display }

    # color
    $viewmenu add cascade -label "Line Color" -menu $viewmenu.color
    menu $viewmenu.color
    $viewmenu.color add radiobutton  -variable color -label " Black " -command {set ccp4i_db_viz(COLOR) black; re_display}
    $viewmenu.color add radiobutton  -variable color -label " Blue " -command {set ccp4i_db_viz(COLOR) blue; re_display}
    $viewmenu.color add radiobutton -variable color -label " Red " -command {set ccp4i_db_viz(COLOR) red; re_display}
    $viewmenu.color add radiobutton -variable color -label " Green " -command {set ccp4i_db_viz(COLOR) green; re_display}

    # direction
    $viewmenu add cascade -label "Orientation" -menu $viewmenu.direction
    menu $viewmenu.direction 

    $viewmenu.direction add radiobutton -variable direction -label "Top-to-buttom" -command  { set ccp4i_db_viz(DIRECTION) TB ; re_display}
    $viewmenu.direction add radiobutton -variable direction -label "Bottom-to-top" -command  {set ccp4i_db_viz(DIRECTION) BT;  re_display }
    $viewmenu.direction add radiobutton -variable direction -label "Left-to-right" -command  { set ccp4i_db_viz(DIRECTION) LR; re_display }
    $viewmenu.direction add radiobutton -variable direction -label "Right-to-left" -command  { set ccp4i_db_viz(DIRECTION) RL; re_display}
    
    pack $bview -side left

    
    ###########
    # Tool bar
    ###########
    set tbar [frame .tbar -background $configure(COLOUR_CONTRAST)]
    set back [button .tbar.back -text "Back"  -command "render_last_view" ]
    set forward [button .tbar.forward -text "Forward" -command "render_forward_view" ]


    set bfilter [button .tbar.filter -text "Apply filter" -command "apply_filter"]

    set bview1 [button .tbar.view1 -text "Show Selected jobs" -command "render_selected_jobs"]
    set bview2 [button .tbar.view2 -text "Show All jobs" -command "render_all_jobs"]
    set bexit [button .tbar.exit -text "Exit" -command "clear"]
    
    bind $back <Enter> "balloonhelp_show $back \"Back to last selection\" "
    bind $back <Leave> "balloonhelp_cancel"

    bind $forward <Enter> "balloonhelp_show $forward \"Forward\" "
    bind $forward <Leave> "balloonhelp_cancel"
   
    bind $bfilter <Enter> "balloonhelp_show $bfilter \"Apply the filter\""
    bind $bfilter <Leave> "balloonhelp_cancel"

    bind $bview1 <Enter> "balloonhelp_show $bview1 \"Show only selected jobs\""
    bind $bview1 <Leave> "balloonhelp_cancel"

    bind $bview2 <Enter> "balloonhelp_show $bview2 \"Show all jobs\""
    bind $bview2 <Leave> "balloonhelp_cancel"
    
    bind $bexit <Enter> "balloonhelp_show $bexit \"Exit the program\""
    bind $bexit <Leave> "balloonhelp_cancel"
    
    pack $back $forward $bfilter $bview1 $bview2 $bexit -side left 

    ###########################
    # Information message line
    ###########################
    set info [label $finfo.label -text "Message line"]
    pack $info -side left
 
    # Pack everything so that it is visible
    pack $fmbar $tbar $f $finfo -side top -fill x
   
    pack $f1 $f2 -side left -fill both -expand true
    pack $c $scrollx -side top -fill x
    pack $scrolly -side right -fill y 
  
  
    # Bind the canvas to scroll when clicked and dragged
    bind $c <ButtonPress-1> "drag_canvas_start $c %X %Y"
    bind $c <ButtonRelease-1> "drag_canvas_cancel $c"

}

#----------------------------------------------------------------
proc render_selected_jobs { } {
#----------------------------------------------------------------
    global ccp4i_db_viz
    
    set g $ccp4i_db_viz(GRAPH)
    set c $ccp4i_db_viz(CANVAS)
    set jobs [return_selection]
    if { [llength $jobs] == 0 } {
	# Nothing to show
	return
    }

    clear_selection $c
    $c delete all
    dotgraph_delete $g
    # remember the last joblist for backwards operation
    lappend ccp4i_db_viz(LAST_SELECTION_LIST) $ccp4i_db_viz(JOB_LIST)
    # reset the filter to 0
    set ccp4i_db_viz(FINISHED_FILTER) 0
    set ccp4i_db_viz(LINKED_FILTER) 0
    render_jobs $jobs 
}

proc apply_filter {} {
   
    global ccp4i_db_viz

    # check if filter is on
    if { $ccp4i_db_viz(FINISHED_FILTER) == 0  && $ccp4i_db_viz(LINKED_FILTER) == 0 } {
	WarningMessage  " no filter set "
	return 
    }

    # clear the canvas
    set g $ccp4i_db_viz(GRAPH)
    set c $ccp4i_db_viz(CANVAS)

    $c delete all
    dotgraph_delete $g

    # set filtered_jobs
    set filtered_jobs {}

    # get selection jobs
    # if there is no selection exist, set current joblist is default selection
    if { $ccp4i_db_viz(SELECTION) == {} } {
	set ccp4i_db_viz(SELECTION) $ccp4i_db_viz(JOB_LIST)
    }

    # check the filter
    if { $ccp4i_db_viz(FINISHED_FILTER) == 1 } {
	set filtered_jobs [ get_finished_jobs $ccp4i_db_viz(SELECTION) ] 
      
    }
    if { $ccp4i_db_viz(LINKED_FILTER) == 1 } {
	if { $filtered_jobs == {} } {
	    set filtered_jobs [ get_linked_jobs $ccp4i_db_viz(SELECTION) ]
	} else {
	    set filtered_jobs [ get_linked_jobs $filtered_jobs ]
	}
    }
    update_info
    render_jobs $filtered_jobs
    
}
#---------------------------------------------------------------
proc get_finished_jobs { joblist } {
#---------------------------------------------------------------
# get finished jobs 
    
    global ccp4i_db_viz
    
    # get finished jobs from handler
    set finishedjobs [ SelectJobs $ccp4i_db_viz(PROJECT) STATUS {FINISHED} ]

    # filter the id in the current SELECTION, if it is a finished job, applend to the temp job list
    foreach id $joblist {
	if { [lsearch $finishedjobs $id] > -1 } {
	    lappend filteredjobs $id }
    }
    return $filteredjobs
}

#---------------------------------------------------------------
proc get_linked_jobs { joblist } {
#---------------------------------------------------------------
# render linked job only
  
    global ccp4i_db_viz

    # get the all link in joblist
    set connection [GetLinks $ccp4i_db_viz(PROJECT) $joblist]

    # filter id in joblist
    foreach id $joblist {
	foreach link $connection {
	    if { [lsearch $link $id] >-1 } {
		lappend filteredjobs $id
		break 
	    } 
     }
 }
    return $filteredjobs
}

#----------------------------------------------------------------
proc render_last_view { } {
#----------------------------------------------------------------
# display the last view

    global ccp4i_db_viz
    set g $ccp4i_db_viz(GRAPH)
    set c $ccp4i_db_viz(CANVAS)
    
    set l [llength $ccp4i_db_viz(LAST_SELECTION_LIST)]
    if { $l == 0 } {
	return }
    set jobs [lindex $ccp4i_db_viz(LAST_SELECTION_LIST) [expr ($l-1)]]

    # need to remove the last one after display it
    set ccp4i_db_viz(LAST_SELECTION_LIST) [lrange $ccp4i_db_viz(LAST_SELECTION_LIST) 0 [expr ($l-2)]]
   
    # add to the forward_selection_list
    lappend ccp4i_db_viz(FORWARD_SELECTION_LIST) $ccp4i_db_viz(JOB_LIST)

    # clear the canvas
    clear_selection $c
    $c delete all
    dotgraph_delete $g
    render_jobs $jobs  

}

#----------------------------------------------------------------
proc render_forward_view { } {
#----------------------------------------------------------------
# display forward view

    global ccp4i_db_viz
    set g $ccp4i_db_viz(GRAPH)
    set c $ccp4i_db_viz(CANVAS)
    set l [llength $ccp4i_db_viz(FORWARD_SELECTION_LIST)]
    if { $l == 0 } {
	 return }
    # get the last element in the forward_selection_list
    set jobs [lindex $ccp4i_db_viz(FORWARD_SELECTION_LIST) [expr ($l-1)]]

    # need to remove the last one after display it
    set ccp4i_db_viz(FORWARD_SELECTION_LIST) [lrange $ccp4i_db_viz(FORWARD_SELECTION_LIST) 0 [expr ($l-2)]]
    
    # add to the last_selection_list
    lappend ccp4i_db_viz(LAST_SELECTION_LIST) $ccp4i_db_viz(JOB_LIST)

    # clear the canvas
    clear_selection $c
    $c delete all
    dotgraph_delete $g
    render_jobs $jobs  
    
}
#----------------------------------------------------------------
proc render_all_jobs { } {
#----------------------------------------------------------------
    global ccp4i_db_viz
  
    set g $ccp4i_db_viz(GRAPH)
    set c $ccp4i_db_viz(CANVAS)
    set jobs [ListJobs $ccp4i_db_viz(PROJECT)]
    if { [llength $jobs] == 0 } {
	# Nothing to show
	return
    }
    clear_selection $c
    $c delete all
    dotgraph_delete $g
    # reset filter to 0
    set ccp4i_db_viz(FINISHED_FILTER) 0
    set ccp4i_db_viz(LINKED_FILTER) 0
    render_jobs $jobs
}


#----------------------------------------------------------------
proc render_jobs { jobs } {
#----------------------------------------------------------------
    # Create the display of the list of jobs
    # "jobs" is a Tcl list of job ids supplied by the caller
    global ccp4i_db_viz
  
    # "configure is CCP4i's global configuration
    # array which contains font and colour information
    global configure

    # Initialise parameters for display
    set maxwidth $ccp4i_db_viz(MAXWIDTH)
    set maxheight $ccp4i_db_viz(MAXHEIGHT)
    set minwidth $ccp4i_db_viz(MINWIDTH)
    set minheight $ccp4i_db_viz(MINHEIGHT)
    set border $ccp4i_db_viz(BORDER)

    # Collect canvas name
    set c $ccp4i_db_viz(CANVAS)
  
    # create a dotgraph
    set g [ dotgraph_new digraph -direction $ccp4i_db_viz(DIRECTION) -background $ccp4i_db_viz(BACKGROUND) -shape $ccp4i_db_viz(SHAPE) -color $ccp4i_db_viz(COLOR) -fillcolor $ccp4i_db_viz(FILLCOLOR) ]

    # Store the name of the graph for later
    set ccp4i_db_viz(GRAPH) $g

    # Write out the nodes (= jobs)
    foreach id $jobs {
	    
	    # Set the colour depending on the status     
	    set status [getdata $id STATUS]
	    set colour [get_job_colour $status]
	    
	    # Create a node for this job
	    set name "Job$id"
	    set taskname [getdata $id TASKNAME]
	    dotgraph_add_node $g $name \
		-label "$id: $taskname" \
		-shape box -fillcolor $colour
	    
	    ####################################
	    # Add bindings for node
	    ####################################
	    
	    # Bindings that are invoked when the mouse enters and leaves
	    # a displayed job
	    #dotgraph_bind_node $g $name <Enter> "display_job_data $id"
       
	    dotgraph_bind_node $g $name <Enter> "bubblehelp_for_job $id %X %Y %x %y"
	    #dotgraph_bind_node $g $name <Leave> "clear_job_data"
	    dotgraph_bind_node $g $name <Leave> "bubblehelp_for_job_cancel "
	    # Selection
	    # Ctrl-Button1 selects multiple jobs
	    dotgraph_bind_node $g $name <Control-Button-1> "select_multiple_jobs $c $id"
	    # Button1 selects a single job
	    dotgraph_bind_node $g $name <Button-1> "select_job $c $id"
	    
	    # Right mouse click on a job to give a menu of options
	    dotgraph_bind_node $g $name <Button-3> "display_job_options $c %X %Y $id"
	}

    set connection [GetLinks $ccp4i_db_viz(PROJECT) $jobs]
 
    foreach link $connection {
	set id_1 [lindex $link 0]
	set id_2 [lindex $link 1]

	# check the id in link also in jobs list
	if { [lsearch $jobs $id_1]> -1 && [lsearch $jobs $id_2] >-1 } {
	    set node1 "Job$id_1"
	    set node2 "Job$id_2"

	    dotgraph_add_connection $g $node1 $node2 }
    } 
    
    # Display the graph
    dotgraph_render $g $c
    
    # Reset the height and width of the canvas
    set bbox [$c bbox all]
    
    set new_height [expr [lindex $bbox 3] - [lindex $bbox 1] + 2 * $border]
    if { $new_height > $maxheight } {
	set new_height $maxheight
    } elseif { $new_height < $minheight } {
	set new_height $minheight
    }
    $c configure -height $new_height
    
    set new_width [expr [lindex $bbox 2] - [lindex $bbox 0] + 2 * $border]
    if { $new_width > $maxwidth } {
	set new_width $maxwidth
    } elseif { $new_width < $minwidth } {
	set new_width $minwidth
    }
    $c configure -width $new_width
    
    # Reset the maximum scrollable region
    set bbox [$c bbox all]
    $c configure -scrollregion [list \
				    [expr [lindex $bbox 0] - $border] \
				    [expr [lindex $bbox 1] - $border] \
				    [expr [lindex $bbox 2] + $border] \
				    [expr [lindex $bbox 3] + $border]]
    
    # Reset the view to the top left-hand corner
    $c yview moveto 0
    $c xview moveto 0

    # Store the list of rendered jobs
    set ccp4i_db_viz(JOB_LIST) $jobs
}

#----------------------------------------------------------------
proc get_job_colour { status } {
#----------------------------------------------------------------
    # Set the colour for the job depending on its status, Default is gray
    set colour "gray"
    switch $status {
	FINISHED { set colour green }
	FAILED { set colour red }
	KILLED { set colour blue }
	RUNNING { set colour yellow }
    }
    return $colour
}

#---------------------------------------------------------------
proc output_graph { } {
#---------------------------------------------------------------
    global ccp4i_db_viz
    
    # if no graph loaded, give warning message
    if { $ccp4i_db_viz(GRAPH) == {} } {
	WarningMessage "No graph to print"
	return }

    # Open another window
    set w .anysection
    if ![OpenWindow $w "Print Graph" "Print" ] { return }

    # Make the frame and the cancel and apply buttons
    CreateFrame   $w output_graph
    CreateButton  $w dismiss "Cancel" "unset output_graph ; destroy $w"
    CreateButton  $w output  "Output" "apply_output_graph output_graph"

    # Define the menus used for this window
    DefineMenu _viewer_format [list "JPEG" "GIF" "Postscript" "PNG"] \
	[list "jpg" "gif" "ps" "png"]

    # Set parameters to change
    DefineVariable output_graph FORMAT _viewer_format "ps"
    DefineVariable output_graph OUTPUT_FILE   _any_file          ""
    DefineVariable output_graph DIR_OUTPUT_FILE   _default_dirs     ""

    # Draw the widgets in the window
    CreateLine line \
	label "Output graph to file as" \
	widget FORMAT
    CreateOutputFileLine line "Name of file to write to" \
	"Output file" \
	OUTPUT_FILE DIR_OUTPUT_FILE
    CloseFrame

}


#--------------------------------------------------------------------
proc apply_output_graph { outputVar } {
#--------------------------------------------------------------------
    upvar \#0 $outputVar output_graph 
   
    global ccp4i_db_viz
 
    # Get the format & file name
    set format [GetValue $outputVar FORMAT]
    set name [GetFullFileName0 $outputVar OUTPUT_FILE]

    if { [file exists $name ]} {
      
	set action [ ChooseOptionDialog "File Exists" "File Exists" \
			 "File exists: $name \nThis file will be overwritten. Do you want to continue or abort the output operation?" \
			 [list "Cancel" "Continue"]  \
			 -default 0 ]
	if {"$action"=="Cancel"} {
	    #WarningMessage "Graph output cancelled."
	    return
	}
    }
	
    if { ! [dotgraph_render_output $ccp4i_db_viz(GRAPH) $format $name] } {
	WarningMessage "print file failed"
	return
    } else {
	# check the file create
	if { [file exists $name] } { 
	    showmessage "a file $name has been output"
	    destroy .anysection 
	} else {
	    WarningMessage "print file faied"
	    return
	}
    }
}

proc update_info { } {
    global ccp4i_db_viz

    # check the filter state
    if { $ccp4i_db_viz(FINISHED_FILTER) == 1 } {
	set F1 "ON"
    } else {
	set F1 "OFF"
}

    if { $ccp4i_db_viz(LINKED_FILTER) == 1 } {
	set F2 "ON"
s    } else {
	set F2 "OFF"
    }
    
    .finfo.label configure \
	    -text "Project: $ccp4i_db_viz(PROJECT)   Finished filter: $F1  Linked filter: $F2"
}
############################################################
# Procedures for moving the canvas by doing "click and drag"
############################################################

#----------------------------------------------------------------
proc drag_canvas_start { c start_x start_y } {
#----------------------------------------------------------------
    # Invoked when user clicks on the canvas
    # Any dragging motion will trigger the next binding, to scroll
    # the canvas
    bind $c <Motion> "drag_canvas_scroll $c $start_x $start_y %X %Y"
    # Change cursor to a cross - this indicates to
    # the user that the function of the mouse has changed
    $c config -cursor fleur
}

#----------------------------------------------------------------
proc drag_canvas_scroll { c start_x start_y x y } {
#----------------------------------------------------------------
    # Invoked following click-drag operation on the canvas
    # Scroll the viewable region as the mouse is dragged around
    $c xview scroll [expr $x - $start_x] "units"
    $c yview scroll [expr $y - $start_y] "units"
    bind $c <Motion> "drag_canvas_scroll $c $x $y %X %y"
}


#----------------------------------------------------------------
proc drag_canvas_cancel { c } {
#----------------------------------------------------------------
    # Invoked when the mouse button is released
    # Cancels the scrolling operation on the canvas view
    bind $c <Motion> {}
    # Change cursor back to default
    $c config -cursor {} 
}


############################################################
# Procedures for displaying job information
############################################################

#----------------------------------------------------------------
proc display_job_data { id } {
#----------------------------------------------------------------
    # Placeholder: invoked when mouse enters node
    # In a full application this could display the
    # job information e.g. in a separate panel or as ballon
    # information
    global ccp4i_db_viz
    .finfo.label configure \
	-text "Job $id: [getdata $id TASKNAME] \"[getdata $id TITLE]\""
}

#----------------------------------------------------------------
proc clear_job_data { } {
#----------------------------------------------------------------
    # Remove the job data displayed in the line at
    # the bottom of the screen
    .finfo.label configure -text {}
}


############################################################
# Procedures for displaying options on a right mouse click
############################################################

#----------------------------------------------------------------
proc display_job_options { c x y id } {
#----------------------------------------------------------------
    # Placeholder: invoked on right mouse click
    # In a full application this could give options to
    # rerun, view files, delete etc - essentially
    # any manipulation operations that could be
    # performed on a single job
    # Make a pop-up menu
    set m "$c.popup"
    if { [winfo exists $m] } {
	destroy $m
    }
    menu $m -tearoff 0
   
    $m add command -label "Select all children" \
	-command "select_all_children $c $id"
    $m add command -label "Select all parents" \
	-command "select_all_parents $c $id"
    # Pop-up the menu
    tk_popup $m $x $y
}



############################################################
# Procedures for selecting/deselecting jobs
############################################################

#----------------------------------------------------------------
proc select_job { c id } {
#----------------------------------------------------------------
    # Invoked on left mouse click to perform job selection
    #
    # If the job is not already selected then deselect
    # any existing selection and then select this job
    # If the job is selected then deselect it and all others
    global ccp4i_db_viz
    # Check that this job is not already in the selection
    if { [set i [lsearch $ccp4i_db_viz(SELECTION) $id]] < 0 } {
	# Not found = not selected
	# Clear existing selection
	clear_selection $c
	# Add this job
	lappend ccp4i_db_viz(SELECTION) $id
       
	# Change the colour
	update_colour $c $id "cyan"
    } else {
	# Job is already selected - so remove it
	# and all others
	clear_selection $c
    }
}

#----------------------------------------------------------------
proc select_multiple_jobs { c id } {
#----------------------------------------------------------------
    # Invoked on left mouse click with control to add
    # or remove a job to/from a multiple job selection
    #
    # If the job is not already selected then add it
    # to any existing selection
    # If the job is selected then deselect it
    global ccp4i_db_viz
   
    # Check that this job is not already in the selection
    if { [set i [lsearch $ccp4i_db_viz(SELECTION) $id]] < 0 } {
	# Not found = not selected
	# Add this job
	lappend ccp4i_db_viz(SELECTION) $id
 
	# Change the colour
	update_colour $c $id "cyan"
    } else {
	# Job is already selected - so remove it
	# from the selection
	set status [getdata $id STATUS]
	set colour [get_job_colour $status]
	# Configure the item
	update_colour $c $id $colour
	# Remove from the selection list
	set ccp4i_db_viz(SELECTION) [lreplace $ccp4i_db_viz(SELECTION) $i $i]
    }
}

#--------------------------------------------------------------------
proc select_all_children { c id { no_clear 0 } } {
#--------------------------------------------------------------------
# select all the children of a specific job

    global ccp4i_db_viz
    # Clear the existing selection
    if { ! $no_clear } {
	clear_selection $c
    }

    # Select the current job (if not already selected)
    if { [lsearch $ccp4i_db_viz(SELECTION) $id] < 0 } {
	lappend ccp4i_db_viz(SELECTION) $id
    }
    update_colour $c $id "cyan"

    set children [GetAllChildren $ccp4i_db_viz(PROJECT) $id]
   
    foreach child $children {
	update_colour $c $child "cyan"
	lappend ccp4i_db_viz(SELECTION) $child }

}


#--------------------------------------------------------------------
proc select_all_parents { c id { no_clear 0 } } {
#--------------------------------------------------------------------
# select all the parents of a specific job

    global ccp4i_db_viz
    # Clear the existing selection
    if { ! $no_clear } {
	clear_selection $c
    }

    # Select the current job (if not already selected)
    if { [lsearch $ccp4i_db_viz(SELECTION) $id] < 0 } {
	lappend ccp4i_db_viz(SELECTION) $id
    }
    update_colour $c $id "cyan"

    set parents [GetAllParents $ccp4i_db_viz(PROJECT) $id]
   
    foreach parent $parents {
	update_colour $c $parent "cyan"
	lappend ccp4i_db_viz(SELECTION) $parent }

}


#----------------------------------------------------------------
proc return_selection { } {
#----------------------------------------------------------------
    # Return the current list of selected jobs
    global ccp4i_db_viz
    return $ccp4i_db_viz(SELECTION)
}


#----------------------------------------------------------------
proc clear_selection { c } {
#----------------------------------------------------------------
    # Clear the list of selected jobs
    global ccp4i_db_viz
    
    # Loop over all currently selected job ids
    # and restore the colour
    foreach id $ccp4i_db_viz(SELECTION) {
	# Set the colour according to the status
	set colour [get_job_colour [getdata $id STATUS]]
	# Configure the item
	update_colour $c $id $colour
    }
    # Erase the list of jobs
    set ccp4i_db_viz(SELECTION) {}
    return $ccp4i_db_viz(SELECTION)
}


#----------------------------------------------------------------
proc update_colour { c id newcolor } {
#----------------------------------------------------------------
    # Change the colour of a node on the fly to the selected
    # colour
    global ccp4i_db_viz
    set g $ccp4i_db_viz(GRAPH)
    dotgraph_update_node_color $g "Job$id" $newcolor $c
}


############################################################
# Procedures for display information about the links
############################################################

#----------------------------------------------------------------
proc display_link_data { id1 id2 } {
#----------------------------------------------------------------
    # Placeholder: invoked when mouse moves over link
    puts "Link between jobs $id1 and $id2"
}

############################################################
# Procedures for displaying "bubble help"
############################################################
#

#----------------------------------------------------------------
proc show_info { c tag id x y } {
#----------------------------------------------------------------
    global ccp4i_db_viz
    $c bind $tag <Leave> "cancel_info $c $tag $id"
    # Set up the string to be displayed
    set label  "[getdata $id TITLE]"
    # Display the text
    .nodeinfo.label configure -text "$label"
    wm geometry .nodeinfo +$x+$y
    wm deiconify .nodeinfo
    raise .nodeinfo
}


#----------------------------------------------------------------
proc cancel_info { c tag id } {
#----------------------------------------------------------------
    $c bind $tag <Enter> "show_info $c $tag $id %X %Y"
    wm withdraw .nodeinfo
    .nodeinfo.label configure -text ""
}

#----------------------------------------------------------------
proc bubblehelp_for_job { id X Y x1 y1 } {
#----------------------------------------------------------------
# display bubblehelp information for specific job
    global ccp4i_db_viz
    set w $ccp4i_db_viz(CANVAS)

    # Get the node name based on the job id
    # This is the tag created by dotgraph and used for the
    # canvas item representing the job
    set name "Job$id"

    # Generate the text to display and load into the label
    set title [getdata $id TITLE]
    set date [getdata $id DATE]
    set re_format_date [GetDate -format brief -clock $date]
    set label "Job $id: \nDate: $re_format_date\nTitle: $title\n"

    append label "Input files:"
    if { [llength [getdata $id INPUT_FILES]] > 0 } {
	foreach filen [getdata $id INPUT_FILES] {
	    append label "\n    * $filen"
	}
    } else {
	append label "\n  <None>"
    }
    append label "\nOutput files:"
    if { [llength [getdata $id OUTPUT_FILES]] > 0 } {
	foreach filen [getdata $id INPUT_FILES] {
	    append label "\n    * $filen"
	}
    } else {
	append label "\n    <None>"
    }
    .nodeinfo.label configure -text "$label" -justify left

    # Calculate the position to display the bubble on the screen
    # This is not straightforward because we have three different
    # coordinate systems:
    # * X and Y are the absolute screen coordinates
    # * x1 and y1 are coordinates relative to the *viewport* origin
    # * canvasx and canvasy translate the viewport coordinates into
    #   coordinates relative to the *canvas* origin
    # * bbox command gives the bounding box of the items relative to
    #   the canvas origin
    #
    # So the procedure is:
    # 1. Get the mouse position in the canvas, by transforming x1,y1
    #    into x2,y2 using the canvasx and canvasy commands
    # 2. Get the coordinates of the box around the node (also in
    #    canvas coordinates) using the bbox command
    # 3. Get the coordinates of the bottom left-hand corner of the
    #    bounding box in canvas coordinates, as x3,y3
    # 4. Calculate the shift from the mouse position x2,y3 to the
    #    corner of the box x3,y3
    # 5. Apply this shift (plus some extra) to the mouse position in
    #    absolute coordinates to get the screen position of the
    #    bubble
    
    # Now do it for real:

    # Get the mouse position in the canvas coordinate system
    # Turn these values into integers
    set x2 [string trim [format %-6.0f [$w canvasx $x1]]]
    set y2 [string trim [format %-6.0f [$w canvasy $y1]]]

    # Get the bounding box coordinates of the current node
    # These are in the same coordinate frame as x2,y2
    set bbox [$w bbox $name]

    # Get the bottom LH corner of the box as x3,y3
    set x3 [lindex $bbox 0]
    set y3 [lindex $bbox 3]

    # Calculate shifts from x2,y2 to x3,y3
    set xshift [expr $x3 - $x2]
    set yshift [expr $y3 - $y2]
    
    # Apply shifts to the absolute screen coordinates X,Y (plus
    # some extra padding)
    set padding 5
    set x [expr $X + $xshift + $padding]
    set y [expr $Y + $yshift + $padding]

    # Move the bubblehelp to this position and make visible
    wm geometry .nodeinfo +$x+$y
    wm deiconify .nodeinfo
    raise .nodeinfo   
}

#----------------------------------------------------------------
proc bubblehelp_for_job_cancel { } {
#----------------------------------------------------------------
# destroy bubblehelp for specific job

     wm withdraw .nodeinfo
    .nodeinfo.label configure -text ""
    
}

#----------------------------------------------------------------
#proc balloonhelp_for { win mesg } {
#----------------------------------------------------------------
#     
#    global bhInfo
#    set bhInfo($win) $mesg
#    bind $win <Enter> {balloonhelp_show %w}
#    bind $win <Leave> {balloonhelp_cancel}
#}

#----------------------------------------------------------------
#proc balloonhelp_pending {win} {
#----------------------------------------------------------------
#
#    global bhInfo
#    balloonhelp_cancel
#    set bhInfo(pending) [after 1500 [list balloonhelp_show $win]]
#}

#-----------------------------------------------------------------
proc balloonhelp_cancel { } {
#-----------------------------------------------------------------
    global bhInfo
   # if {[info exists bhInfo(pending)]} {
#	after cancel $bhInfo(pending)
#	unset bhInfo(pending)
#    }
    wm withdraw .balloonhelp
}

#-----------------------------------------------------------------
proc balloonhelp_show {win msg} {
#-----------------------------------------------------------------
    #global bhInfo
   
    .balloonhelp.info configure -text $msg
    set  x [expr [winfo rootx $win]+10]
    set y [expr [winfo rooty $win]+[winfo height $win]]
    wm geometry .balloonhelp +$x+$y
    wm deiconify .balloonhelp
    raise .balloonhelp
    
    #unset bhInfo(pending)
}

proc select_project { current_project } {
#-----------------------------------------------------------------
# procedures when select a project 

    global ccp4i_db_viz
    
    if { $current_project == $ccp4i_db_viz(PROJECT) } {
	return
    } else {
	# close previous project
	if { $ccp4i_db_viz(PROJECT) != "" } {
	    CloseDatabase $ccp4i_db_viz(PROJECT)}

	    # delete dotgraph,canvas if it exists	
	    set g $ccp4i_db_viz(GRAPH)
	    set c $ccp4i_db_viz(CANVAS)
	
	    if { $c != "" } {
	    clear_selection $c
	    $c delete all
	}
       	    if { $g != "" } {
		dotgraph_delete $g
	    }
	 # display project
	    set ccp4i_db_viz(PROJECT) $current_project
	    display_project $current_project
    }
}


#------------------------------------------------------------------
proc refresh { } {
#------------------------------------------------------------------
# refresh current project display when receive  broadcast

     global ccp4i_db_viz
     # delete dotgraph,canvas
     set g $ccp4i_db_viz(GRAPH)
     set c $ccp4i_db_viz(CANVAS)
     clear_selection $c
     $c delete all
     dotgraph_delete $g
   
     set itemlist {STATUS TASKNAME TITLE DATE INPUT_FILES OUTPUT_FILES}
    
    #getdata from handler
     set jobs [ListJobs $ccp4i_db_viz(PROJECT)]
    
     if { $jobs != 0 && $jobs != []} {
	 GetDbData $ccp4i_db_viz(PROJECT) $jobs $itemlist
	
	 #Render the job list
	 render_jobs $jobs

     } else {
	 WarningMessage "No jobs in the project"
     }
 }

#------------------------------------------------------------------
proc re_display { } {
#------------------------------------------------------------------
# re-display when re-set the direction, background, color etc.

    global ccp4i_db_viz
  
    if { $ccp4i_db_viz(JOB_LIST) != {} } {
	# delete dotgraph,canvas
	set g $ccp4i_db_viz(GRAPH)
	set c $ccp4i_db_viz(CANVAS)
	clear_selection $c
	$c delete all
	dotgraph_delete $g
	render_jobs $ccp4i_db_viz(JOB_LIST)
    } else {
	WarningMessage "no job to display"
    }
}

#------------------------------------------------------------------
proc GetDbData { project joblist itemlist} {
#------------------------------------------------------------------
# procedure to get the data from handler and stort in database array

    global database 
    # get a list records from handler
    set records [GetListofRecords $project $joblist $itemlist ]

    # store the records in database array
    foreach id $joblist {
	# get index of item
	set i [lsearch $joblist $id]

	# get record for the jobid 
	set record [lindex $records $i]
	foreach item $itemlist {
	    # get index of id
	    set j [lsearch $itemlist $item]
	    # get the item from the record
	    set database($item,$id) [lindex $record $j]
	}
    }
}

#-----------------------------------------------------------
proc getdata { id item } {
#-----------------------------------------------------------
# procedure to get the data from database array

    global database
    return $database([subst $item],$id)
}

#-----------------------------------------------------------
proc display_project { project } {
#-----------------------------------------------------------
# procedure of open & display a project 
    global ccp4i_db_viz

    # set filter off
    set ccp4i_db_viz(FINISHED_FILTER) 0
    set ccp4i_db_viz(LINKED_FILTER) 0
    #OpenDatabase
    if { [OpenDatabase $project msg] == 0 } {
       
	if { $msg == "data is locked" } {
	set action [ ChooseOptionDialog "Lock Exists" "Lock Exists" \
			 " The database for project $project has a lock. The lock may be left over from other process which did not terminate cleanly. Do you want to overwrite the lock? " \
			 [list "No" "Yes"]  \
			 -default 0 ]
	if {"$action"=="No"} {
	    return
	} else {
          OpenDatabase $project -grablock
	}
	} else {
	    WarningMessage "$msg"   
	    return
	}
    } 
  
    #define what item to get from database
    set itemlist {STATUS TASKNAME TITLE DATE INPUT_FILES OUTPUT_FILES}
    
    #getdata from handler
    set jobs [ListJobs $project]
    
    if { $jobs != 0 && $jobs != []} {
	GetDbData $project $jobs $itemlist
	
	#Render the job list
	render_jobs $jobs

	# store the current project in the global array for later use
	#set ccp4i_db_viz(PROJECT) $project

	# show project name in message line
	#.finfo.label configure \
	    -text "Project: $project"
	update_info 
    } else {
	# unset joblist
	set ccp4i_db_viz(JOB_LIST) {}
	WarningMessage "No jobs in the project"
    }
} 


#-----------------------------------------------------------
proc clear { } {
#-----------------------------------------------------------
    # destroy the root window
    destroy .
    # disconnect handler
    DbHandlerDisconnect   
    exit
}

#-----------------------------------------------------------
proc list_projects { } {
#-----------------------------------------------------------
# List projects and allow user to select by double click one of the projects

    toplevel .projects 
    frame .projects.f -height 25 -width 10
    pack .projects.f -expand yes -fill both -padx 1 -pady 2

    listbox .projects.f.projectlist -yscrollcommand { .projects.f.sbar set } -height 10
    scrollbar .projects.f.sbar -command { .projects.f.projectlist yview }
    pack .projects.f.projectlist .projects.f.sbar -side left -fill both
    set projects [ListProjects]
    foreach project $projects { .projects.f.projectlist insert end $project }
    bind .projects.f.projectlist <Double-Button-1> {
	set current_project [selection get]
	select_project $current_project
    }
}

proc output_xml { } {
    global ccp4i_db_viz
    puts $ccp4i_db_viz(PROJECT)
    set jobs [return_selection]
    puts $jobs
    exec python output_tracking_data.py $ccp4i_db_viz(PROJECT) $jobs

}

#------------------------------------------------------------------
proc showmessage { message } {
#------------------------------------------------------------------
# messagebox
    toplevel .message
    frame .message.f
    pack .message.f -expand yes -fill both -padx 2 -pady 2
    #label .message.f.icon -bitmap error
    #pack .message.f.icon -side left -padx 8 -pady 8
    label .message.f.info -text $message
    pack .message.f.info -side right -expand yes -fill both -padx 8 -pady 8
    frame .message.f2 -height 2 -borderwidth 1 -relief sunken
    pack .message.f2 -fill x -padx 4
    button .message.dismiss -text "Dismiss" -command { destroy .message }
    pack .message.dismiss -pady 4
}

#----------------------------------------------------------------
proc handlebroadcast { broadcast } {
#----------------------------------------------------------------
    # Handle broadcast message from the database handler, refresh the display
        refresh
}

############################################################
# TOP-LEVELRIPT CODE
############################################################

# Start/connect to the database server
if { ![DbStartHandler handlebroadcast] } {
    WarningMessage "Failed to start the handler"
    exit 1
}
if { ![DbHandlerConnect] } {
    WarningMessage "Failed to connect to the handler"
    exit 1
}

# Initialise the dotgraph package
set dotgraph_dir [file join $env(BIOXHIT_TOP) dbccp4i application]
set dot_dir $env(GRAPHVIZ_BIN)

if { ![file exists [file join $dotgraph_dir dotgraph.tcl]] } {
    puts "Failed to locate dotgraph.tcl in $dotgraph_dir"
    exit
}
source [file join $dotgraph_dir dotgraph.tcl]
if { ![dotgraph_init $dot_dir] } {
    puts "Failed to initialise the dotgraph package"
    puts "Maybe you need to reset the variable \"dot_dir\" to point to your"
    puts "local version of the graphviz \"dot\" program?"
    exit
}
if { [dotgraph_version < 0.0.2] } {
    # Needs dotgraph or better 0.0.2
    puts "Needs dotgraph 0.0.2 - this is version [dotgraph_version]"
    exit
}

# Create the main application window
create_main_window

# Info display
toplevel .nodeinfo -borderwidth 1 -background black
label .nodeinfo.label -background white
pack .nodeinfo.label -side left -fill y
wm overrideredirect .nodeinfo 1
wm withdraw .nodeinfo

 #balloonhelp 
toplevel .balloonhelp -borderwidth 1 -background black 
label .balloonhelp.info -background white
pack .balloonhelp.info -side left -fill y
wm overrideredirect .balloonhelp 1
wm withdraw .balloonhelp


#check if there is an initial project given
if { [llength $argv] >0 } {
    set project [lindex $argv 0]
    select_project $project
} else {
    # get the first project
    set projects [ListProjects]
    if { $projects != 0 } {
    set project [ lindex $projects 0 ]
    select_project $project
    } else {
	WarningMessage "No projects"
    }
}

